'use strict';

const isValidXPath = expr => (
    typeof expr !== 'undefined' &&
    expr.replace(/[\s-_=]/g, '') !== '' &&
    expr.length === expr.replace(/[-_\w:.]+\(\)\s*=|=\s*[-_\w:.]+\(\)|\sor\s|\sand\s|\[(?:[^\/\]]+[\/\[]\/?.+)+\]|starts-with\(|\[.*last\(\)\s*[-\+<>=].+\]|number\(\)|not\(|count\(|text\(|first\(|normalize-space|[^\/]following-sibling|concat\(|descendant::|parent::|self::|child::|/gi, '').length
);

const getValidationRegexp = () => {
    let regex =
        "(?P<node>" +
        "(" +
        "^id\\([\"\\']?(?P<idvalue>%(value)s)[\"\\']?\\)" +// special case! `id(idValue)`
        "|" +
        "(?P<nav>//?(?:following-sibling::)?)(?P<tag>%(tag)s)" + //  `//div`
        "(\\[(" +
        "(?P<matched>(?P<mattr>@?%(attribute)s=[\"\\'](?P<mvalue>%(value)s))[\"\\']" + // `[@id="well"]` supported and `[text()="yes"]` is not
        "|" +
        "(?P<contained>contains\\((?P<cattr>@?%(attribute)s,\\s*[\"\\'](?P<cvalue>%(value)s)[\"\\']\\))" +// `[contains(@id, "bleh")]` supported and `[contains(text(), "some")]` is not
        ")\\])?" +
        "(\\[\\s*(?P<nth>\\d|last\\(\\s*\\))\\s*\\])?" +
        ")" +
        ")";

    const subRegexes = {
        "tag": "([a-zA-Z][a-zA-Z0-9]{0,10}|\\*)",
        "attribute": "[.a-zA-Z_:][-\\w:.]*(\\(\\))?)",
        "value": "\\s*[\\w/:][-/\\w\\s,:;.]*"
    };

    Object.keys(subRegexes).forEach(key => {
        regex = regex.replace(new RegExp('%\\(' + key + '\\)s', 'gi'), subRegexes[key]);
    });

    regex = regex.replace(/\?P<node>|\?P<idvalue>|\?P<nav>|\?P<tag>|\?P<matched>|\?P<mattr>|\?P<mvalue>|\?P<contained>|\?P<cattr>|\?P<cvalue>|\?P<nth>/gi, '');

    return new RegExp(regex, 'gi');
};

const preParseXpath = expr => (
    expr.replace(/contains\s*\(\s*concat\(["']\s+["']\s*,\s*@class\s*,\s*["']\s+["']\)\s*,\s*["']\s+([a-zA-Z0-9-_]+)\s+["']\)/gi, '@class="$1"')
);

const xPathToCss = expr => {
    if (!expr) {
        throw new Error(CONSTANTS.ERROR_MESSAGES.MISSING_XPATH_EXPRESSION);
    }

    expr = preParseXpath(expr);

    if (!isValidXPath(expr)) {
        throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH + ': ' + expr);
    }

    const xPathArr = expr.split('|');
    const validationRegexp = getValidationRegexp();
    const cssSelectors = [];
    let xIndex = 0;

    while (xPathArr[xIndex]) {
        const css = [];
        let position = 0;
        let nodes;

        while (nodes = validationRegexp.exec(xPathArr[xIndex])) {
            let attr;

            if (!nodes && position === 0) {
                throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH + ': ' + expr);
            }

            const match = {
                node: nodes[5],
                idvalue: nodes[12] || nodes[3],
                nav: nodes[4],
                tag: nodes[5],
                matched: nodes[7],
                mattr: nodes[10] || nodes[14],
                mvalue: nodes[12] || nodes[16],
                contained: nodes[13],
                cattr: nodes[14],
                cvalue: nodes[16],
                nth: nodes[18]
            };

            let nav = '';

            if (position !== 0 && match['nav']) {
                if (~match['nav'].indexOf('following-sibling::')) {
                    nav = ' + ';
                } else {
                    nav = (match['nav'] === '//') ? ' ' : ' > ';
                }
            }

            const tag = (match['tag'] === '*') ? '' : (match['tag'] || '');

            if (match['contained']) {
                if (match['cattr'].indexOf('@') === 0) {
                    attr = '[' + match['cattr'].replace(/^@/, '') + '*="' + match['cvalue'] + '"]';
                } else {
                    throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH_ATTRIBUTE + match['cattr']);
                }
            } else {
                if (match['matched']) {
                    switch (match['mattr']) {
                        case '@id':
                            attr = '#' + match['mvalue'].replace(/^\s+|\s+$/, '').replace(/\s/g, '#');
                            break;
                        case '@class':
                            attr = '.' + match['mvalue'].replace(/^\s+|\s+$/, '').replace(/\s/g, '.');
                            break;
                        case 'text()':
                        case '.':
                            throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH_ATTRIBUTE + match['mattr']);
                        default:
                            if (match['mattr'].indexOf('@') !== 0) {
                                throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH_ATTRIBUTE + match['mattr']);
                            }
                            if (match['mvalue'].indexOf(' ') !== -1) {
                                match['mvalue'] = '\"' + match['mvalue'].replace(/^\s+|\s+$/, '') + '\"';
                            }
                            attr = '[' + match['mattr'].replace('@', '') + '="' + match['mvalue'] + '"]';
                            break;
                    }
                } else {
                    if (match['idvalue']) {
                        attr = '#' + match['idvalue'].replace(/\s/, '#');
                    } else {
                        attr = '';
                    }
                }
            }

            let nth = '';

            if (match['nth']) {
                if (match['nth'].indexOf('last') === -1) {
                    if (isNaN(parseInt(match['nth'], 10))) {
                        throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH_ATTRIBUTE + match['nth']);
                    }
                    nth = parseInt(match['nth'], 10) !== 1 ? ':nth-of-type(' + match['nth'] + ')' : ':first-of-type';
                } else {
                    nth = ':last-of-type';
                }
            }

            css.push(nav + tag + attr + nth);
            position++;
        }

        const result = css.join('');

        if (result === '') {
            throw new Error(CONSTANTS.ERROR_MESSAGES.INVALID_OR_UNSUPPORTED_XPATH);
        }

        cssSelectors.push(result);
        xIndex++;
    }

    return cssSelectors.join(', ');
};