'use strict';

const LocatorController = ($scope, $http) => {
    const xpathText = document.getElementById('xpath');
    const cssText = document.getElementById('css');

    const init = () => {
        chrome.storage.local.get(['locator', 'hint', 'saveFeatureEnabled', 'isFrameset'], result => {
            $scope.$applyAsync(() => {
                let locator = {};
                if (result.locator) {
                    locator = result.locator;
                }
                try {
                    locator.cssSelector = xPathToCss(locator.xpath);
                } catch (error) {
                    $scope.saveCssSelector = false;
                    console.debug(`${new Date(Date.now()).toISOString()} -> ${error.message}`);
                    locator.cssSelector = null;
                }
                locator.alias = result.hint.text
                    .replace(/[\. ]/, '_')
                    .replace(/[^\w]/gi, '')
                    .toLowerCase();
                $scope.locator = locator;
                $scope.hint = result.hint;
                $scope.saveFeatureEnabled = result.saveFeatureEnabled;
                $scope.isFrameset = result.isFrameset;
            })
        });
    };

    const saveLocator = (savingLocator, settings) => {
        chrome.storage.local.set({locator: savingLocator});
        $("#overlay").fadeIn(300);

        let headers = settings.isBasicAuth ?
            {
                'Authorization': 'Basic ' + settings.auth
            } :
            {
                'X-API-KEY': settings.auth
            };

        $.ajax({
            url: settings.host + '/api/v1/locators',
            headers: headers,
            method: 'POST',
            contentType: "application/json; charset=utf-8",
            dataType: 'json',
            data: JSON.stringify(savingLocator),
            success: (data, textStatus, jqXHR) => {
                if (navigator.webdriver) {
                    showMessage(CONSTANTS.NOTIFICATION_MESSAGES.LOCATOR_SAVED);
                    $scope.closeDialog(true);
                } else {
                    setTimeout(() => {
                        $("#overlay").fadeOut(300);
                        alert(CONSTANTS.NOTIFICATION_MESSAGES.LOCATOR_SAVED);
                        $scope.closeDialog(true);
                    }, 500);
                }
            },
            error: (jqXHR, textStatus, errorThrown) => {
                if (navigator.webdriver) {
                    switch (jqXHR.status) {
                        case 400:
                            showMessage(jqXHR.responseJSON.message);
                            break;
                        case 401:
                            showMessage(CONSTANTS.NOTIFICATION_MESSAGES.WRONG_CREDENTIALS);
                            break;
                        case -1:
                            showMessage(CONSTANTS.NOTIFICATION_MESSAGES.WRONG_HOST);
                            break;
                        default:
                            showMessage(CONSTANTS.NOTIFICATION_MESSAGES.INTERNAL_SERVER_ERROR);
                            break;
                    }
                } else {
                    setTimeout(() => {
                        $("#overlay").fadeOut(300);
                        switch (jqXHR.status) {
                            case 400:
                                alert(jqXHR.responseJSON.message);
                                break;
                            case 401:
                                alert(CONSTANTS.NOTIFICATION_MESSAGES.WRONG_CREDENTIALS);
                                break;
                            case -1:
                                alert(CONSTANTS.NOTIFICATION_MESSAGES.WRONG_HOST);
                                break;
                            default:
                                alert(CONSTANTS.NOTIFICATION_MESSAGES.INTERNAL_SERVER_ERROR);
                                break;
                        }
                    }, 500);
                }
            },
            complete: () => {
                if (navigator.webdriver) {
                    setTimeout(() => {
                        $("#overlay").fadeOut(300);
                    }, 500);
                }
            }
        });
    };

    const copyToClipboard = element => {
        element.select();
        element.setSelectionRange(0, 99999);
        document.execCommand('copy');
        window.getSelection().removeAllRanges();
    };

    const showMessage = string => {
        chrome.runtime.sendMessage({
            action: CONSTANTS.ACTIONS.SEND_NOTIFICATION,
            id: CONSTANTS.NOTIFICATION_IDS.LOCATOR,
            message: string
        });
    };

    const getSelectorsToSend = saveCssSelector => {
        let {xpath, cssSelector} = $scope.locator;

        if (((xpath !== null && xpath.length > 0) &&
                (cssSelector !== null && cssSelector.length > 0 && saveCssSelector)) ||
            ((xpath === null || xpath.length === 0) &&
                (cssSelector !== null && cssSelector.length > 0))) {
            return {
                xpath: null,
                css: cssSelector
            };
        }

        if (((xpath !== null && xpath.length > 0) &&
                (cssSelector !== null && cssSelector.length > 0 && !saveCssSelector)) ||
            ((cssSelector === null || cssSelector.length === 0) &&
                (xpath !== null && xpath.length > 0))) {
            return {
                xpath: xpath,
                css: null
            };
        }

        return {
            xpath: xpath,
            css: cssSelector
        };
    };

    init();

    $scope.save = () => {
        let locator = Object.assign({}, $scope.locator);
        chrome.storage.local.get('settings', result => {
            if (!result.settings) {
                if (navigator.webdriver) {
                    showMessage(CONSTANTS.NOTIFICATION_MESSAGES.SETTINGS_NOT_FOUND);
                } else {
                    alert(CONSTANTS.NOTIFICATION_MESSAGES.SETTINGS_NOT_FOUND);
                }
                return;
            }
            let savingLocator = Object.assign({}, $scope.locator);
            let selectors = getSelectorsToSend(result.settings.saveCssSelector);

            savingLocator.xpath = selectors.xpath;
            savingLocator.cssSelector = selectors.css;

            saveLocator(savingLocator, result.settings);
        });
    };

    $scope.backToHint = () => {
        chrome.runtime.sendMessage({
            action: CONSTANTS.ACTIONS.LOCATOR_DIALOG_RESULT,
            result: {
                type: CONSTANTS.ACTIONS.BACK_TO_HINT,
                hint: $scope.hint,
                isFrameset: $scope.isFrameset
            }
        });
    };

    $scope.testXpath = () => {
        chrome.runtime.sendMessage({
            action: CONSTANTS.ACTIONS.LOCATOR_DIALOG_RESULT,
            result: {
                type: CONSTANTS.ACTIONS.TEST,
                xpath: $scope.locator.xpath,
                cssSelector: $scope.locator.cssSelector,
                isFrameset: $scope.isFrameset
            }
        });
    };

    $scope.closeDialog = skip => {
        chrome.runtime.sendMessage({
            action: CONSTANTS.ACTIONS.LOCATOR_DIALOG_RESULT,
            result: {
                type: CONSTANTS.ACTIONS.SKIP,
                skip,
                isFrameset: $scope.isFrameset
            }
        });
    };

    $scope.copyXpath = () => {
        copyToClipboard(xpathText);
    };

    $scope.copyCss = () => {
        copyToClipboard(cssText);
    };
};

angular.module('gutenberg', ['ui.bootstrap']).controller('LocatorController', LocatorController);

LocatorController.$inject = ['$scope', '$http'];
