const subject7Gutenberg = {};
let started = false;

subject7Gutenberg.smartXpathEngine = (() => {
    let anchors = [];

    const getSmartXpath = (element, hint, targetDocument) => {
        if (hint.useIdAttribute && element.hasAttribute(CONSTANTS.ATTRIBUTES.ID)) {
            return `//${element.localName}[@id='${element.getAttribute(CONSTANTS.ATTRIBUTES.ID)}']`;
        }

        let xpath = getXpathByPatterns(element, hint, targetDocument);
        if (xpath !== null) {
            return xpath;
        }

        xpath = buildXpathByAnchors(element, hint, targetDocument);

        if (xpath === null) {
            throw new Error(CONSTANTS.ERROR_MESSAGES.XPATH_CAN_NOT_BE_GENERATED);
        } else {
            return xpath;
        }
    };

    const getElementByXpath = (xpath, targetDocument) => {
        let result = targetDocument.evaluate(xpath, targetDocument, null, XPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        if (result.invalidIteratorState) {
            throw {code: 2, message: CONSTANTS.ERROR_MESSAGES.INVALID_XPATH};
        }

        let elements = [];
        let element;
        while ((element = result.iterateNext()) !== null) {
            elements.push(element);
        }

        if (elements.length === 0) {
            throw {code: 1, message: CONSTANTS.ERROR_MESSAGES.ELEMENT_NOT_FOUND_XPATH};
        } else {
            if (elements.length === 1) {
                return elements[0];
            } else {
                throw {code: 0, message: CONSTANTS.ERROR_MESSAGES.XPATH_RETURNS_SEVERAL_ELEMENTS, elements: elements};
            }
        }
    };

    const buildXpathByAnchors = (element, hint, targetDocument) => {
        if (!hint.text) {
            return null;
        }
        anchors = [];
        findChildAnchors(element, hint, 0, [], targetDocument);
        findParentAnchors(element, hint, 1, targetDocument);
        if (anchors.length === 0) {
            return null;
        }
        const sortedAnchors = anchors.sort((a, b) => a.distance - b.distance);
        anchors = sortedAnchors;

        let xpath = buildXpath(element, anchors[0], targetDocument);

        return verifyAndClarifyXPath(xpath, element, targetDocument);
    };

    const getXpathByPatterns = (element, settings, targetDocument) => {
        if (settings.usePatterns !== true) {
            return null;
        }
        for (let i = 0; i < settings.patterns.length; i++) {
            let pattern = settings.patterns[i];
            let xpath = pattern.replace(CONSTANTS.REPLACE_STRINGS.HINT);
            let result;
            try {
                result = getElementByXpath(xpath, targetDocument);
            } catch (error) {
                continue;
            }
            if (result === element) {
                return xpath;
            }
        }
        return null;
    };

    const findChildAnchors = (element, hint, distance, skipNodes, targetDocument) => {
        let anchor = getAnchor(element, hint, targetDocument);
        if (anchor !== null && distance <= hint.maxDistance) {
            anchor.distance = distance;
            anchors.push(anchor);
        }

        for (let i = 0; i < element.childNodes.length; i++) {
            let child = element.childNodes[i];
            if (skipNodes && skipNodes.indexOf(child) !== -1) {
                continue;
            }
            findChildAnchors(child, hint, distance + 1, skipNodes, targetDocument);
        }
    };

    const findParentAnchors = (element, hint, distance, targetDocument) => {
        let parent = element.parentElement;
        if (parent === null || distance >= hint.maxDistance) {
            return;
        }
        findChildAnchors(parent, hint, distance + 1, [element], targetDocument);
        findParentAnchors(parent, hint, distance + 1, targetDocument);
    };

    const getAnchor = (element, hint, targetDocument) => {
        let anchor = null;

        if (element.nodeType === Node.ELEMENT_NODE) {
            anchor = buildAnchorByAttribute(element, hint);
            if (!anchor) {
                anchor = buildAnchorByInnerText(element, hint);
            }
        } else if (element.nodeType === Node.TEXT_NODE && element.data) {
            anchor = buildAnchorForTextNode(element, hint);
        }

        if (!anchor) {
            return null;
        }

        let elementByXpath = verifyAndClarifyXPath(CONSTANTS.XPATH_HELPERS.START_XPATH + anchor.condition, element, targetDocument);
        if (elementByXpath === null) {
            return null;
        }
        return anchor;
    };

    const verifyAndClarifyXPath = (xpath, targetElement, targetDocument) => {
        try {
            if (targetElement === getElementByXpath(xpath, targetDocument)) {
                return xpath;
            }
            return null;
        } catch (error) {
            if (error.code === 0 && error.elements.indexOf(targetElement) !== -1) {
                return `(${xpath})[${error.elements.indexOf(targetElement) + 1}]`;
            }
            return null;
        }
    };

    const buildAnchorByInnerText = (element, hint) => {
        if (!element.innerText) {
            return null;
        }
        let innerText = element.innerText.replace(/\s/g, ' ');
        let sourceInnerText = innerText;
        let hintText = hint.text;

        if (hint.matchCase !== true) {
            innerText = innerText.toLocaleLowerCase();
            hintText = hintText.toLocaleLowerCase();
        }
        if (hint.fullMatch === true && innerText === hintText) {
            return {
                element: element,
                condition: `${element.localName}[text()='${sourceInnerText}']`
            };
        } else {
            if (hint.fullMatch === false && innerText.indexOf(hintText) !== -1) {
                let originValue = sourceInnerText.substr(innerText.indexOf(hintText), hintText.length);
                return {
                    element: element,
                    condition: `${element.localName}[contains(text(),'${originValue}')]`
                };
            }
        }
        return null;
    };

    const buildAnchorForTextNode = (element, hint) => {
        if (!element.data) {
            return null;
        }

        let textOrigin = element.data.replace(/\s/g, ' ');
        let text = textOrigin;
        let hintText = hint.text;

        if (hint.matchCase !== true) {
            text = text.toLocaleLowerCase();
            hintText = hintText.toLocaleLowerCase();
        }
        let anchor = null;
        if (hint.fullMatch === true && text === hintText) {
            anchor = {
                element: element,
                condition: `text()[.='${textOrigin}']`
            };
        } else {
            if (hint.fullMatch === false && text.indexOf(hintText) !== -1) {
                let originValue = textOrigin.substr(text.indexOf(hintText), hintText.length);
                anchor = {
                    element: element,
                    condition: `text()[contains(.,'${originValue}')]`
                };
            }
        }
        return anchor;
    };

    const buildAnchorByAttribute = (element, hint) => {
        for (let attribute of CONSTANTS.HINT.ATTRIBUTES_LIST) {
            if (element.hasAttribute(attribute)) {
                let attributeValue = element.getAttribute(attribute);
                let value = hint.text;
                if (hint.matchCase !== true) {
                    attributeValue = attributeValue.toLocaleLowerCase();
                    value = value.toLocaleLowerCase();
                }
                if (hint.fullMatch === true && attributeValue === value) {
                    return {
                        element: element,
                        condition: `${element.localName}[@${attribute}='${element.getAttribute(attribute)}']`
                    }
                } else {
                    if (hint.fullMatch === false && attributeValue.indexOf(value) !== -1) {
                        let originValue = element.getAttribute(attribute).substr(attributeValue.indexOf(value), value.length);
                        return {
                            element: element,
                            condition: `${element.localName}[contains(@${attribute}, '${originValue}')]`
                        }
                    }
                }
            }
        }
        return null;
    };

    const buildXpath = (element, anchor, targetDocument) => {
        let hintElement = anchor.element;
        let xpath = CONSTANTS.XPATH_HELPERS.START_XPATH + anchor.condition;

        if (element === hintElement) {
            return xpath;
        }

        let commonElement = findCommonNode(element, hintElement);
        let pathToAnchor = getPathHierarchy(commonElement, hintElement);
        let pathToElement = getPathHierarchy(commonElement, element);

        if (pathToAnchor.length > 0) {
            xpath += CONSTANTS.XPATH_HELPERS.NODE_DIVIDER +
                CONSTANTS.XPATH_AXES.ANCESTOR +
                CONSTANTS.XPATH_HELPERS.AXES_NODE_DIVIDER +
                commonElement.localName;
        }

        if (pathToElement.length > 0) {
            let xpathPrefix = xpath +
                CONSTANTS.XPATH_HELPERS.NODE_DIVIDER +
                CONSTANTS.XPATH_AXES.DESCENDANT +
                CONSTANTS.XPATH_HELPERS.AXES_NODE_DIVIDER;
            let descendantHierarchy = [];
            for (let currentElement of pathToElement) {
                descendantHierarchy.push(currentElement.localName);
                let currentXpath = xpathPrefix + descendantHierarchy.slice(0).reverse().join(CONSTANTS.XPATH_HELPERS.NODE_DIVIDER);
                try {
                    getElementByXpath(currentXpath, targetDocument);
                } catch (error) {
                    if (error.code === 0 && error.elements.indexOf(element) !== -1 && error.elements.length <= 10) {
                        return currentXpath;
                    } else {
                        continue;
                    }
                }
                return currentXpath;
            }
            return xpath + CONSTANTS.XPATH_HELPERS.NODE_DIVIDER +
                pathToElement
                    .reverse()
                    .map(node => node.localName)
                    .join(CONSTANTS.XPATH_HELPERS.NODE_DIVIDER);
        } else {
            return xpath;
        }
    };

    const findCommonNode = (node1, node2) => {
        let parentNode1 = node1;
        while (parentNode1.parentNode !== null) {
            let parentNode2 = node2;
            while (parentNode2.parentNode !== null) {
                if (parentNode1 === parentNode2) {
                    return parentNode1;
                }
                parentNode2 = parentNode2.parentNode;
            }
            parentNode1 = parentNode1.parentNode;
        }
        return null;
    };

    const getPathHierarchy = (fromElement, toElement) => {
        let currentElement = toElement;
        let path = [];
        while (currentElement !== null && currentElement !== fromElement) {
            path.push(currentElement);
            currentElement = currentElement.parentNode
        }
        return path;
    };

    return {
        generateSmartXPath: getSmartXpath,
        getElementByXpath: getElementByXpath,
    };
})();

subject7Gutenberg.cssSelectorEngine = (() => {
    const generateCssForElementInShadowDom = (composedPath) => {
        if (!composedPath) {
            throw new Error(CONSTANTS.ERROR_MESSAGES.CSS_CAN_NOT_BE_GENERATED);
        }

        let cssSelectorArr = [];

        composedPath
            .reverse()
            .forEach((element, index, array) => {
                if (element.nodeType !== Node.DOCUMENT_FRAGMENT_NODE) {
                    let item = getElementSelector(element);

                    if (subject7Gutenberg.locatorRecorder.getOpenOrClosedShadowRoot(element) !== null) {
                        cssSelectorArr.push(
                            array[index + 1] && array[index + 1].nodeType === Node.DOCUMENT_FRAGMENT_NODE ?
                                item.concat(CONSTANTS.CSS.SHADOW_PSEUDO) :
                                item.concat(CONSTANTS.CSS.SHADOW_PSEUDO_CLOSED)
                        );
                    } else {
                        cssSelectorArr.push(item);
                    }
                }
            });

        return cssSelectorArr.join(' ');
    };

    const getElementSelector = element => {
        let selector = getTagName(element);

        if (element.id) {
            selector += '#'.concat(element.id);
            return selector;
        }

        let classes = element.hasAttribute(CONSTANTS.ATTRIBUTES.CLASS) ?
            element.getAttribute(CONSTANTS.ATTRIBUTES.CLASS).trim() :
            '';
        if (classes.length > 0) {
            selector += '.'.concat(classes.replaceAll(' ', '.'));
        }

        let elArr = [...element.parentNode.querySelectorAll(selector)]
        let index = elArr.indexOf(element);
        if (elArr.length > 1) {
            selector += CONSTANTS.CSS_HELPERS.NTH_OF_TYPE
                .replace(CONSTANTS.REPLACE_STRINGS.VALUE, String(index + 1));
        }

        return selector;
    };

    const getTagName = el => {
        return el.tagName.toLowerCase();
    };

    const getNodeName = el => {
        return el.nodeName.toLowerCase();
    };

    const getElementByCss = (css, targetDocument) => {
        let elements;
        try {
            elements = css.includes(CONSTANTS.CSS.SHADOW_PSEUDO) ?
                findElementInShadowDom(css, targetDocument) :
                targetDocument.querySelectorAll(css);
        } catch (e) {
            switch (e.message) {
                case CONSTANTS.ERROR_MESSAGES.SHADOW_CSS_CANNOT_END_ON_A_SHADOW_ROOT:
                    throw {code: 2, message: e.message};
                case CONSTANTS.ERROR_MESSAGES.ELEMENT_NOT_FOUND_CSS:
                    throw {code: 1, message: e.message};
                default:
                    throw {code: 2, message: CONSTANTS.ERROR_MESSAGES.INVALID_CSS.concat(e.message)};
            }
        }

        if (elements.length === 0) {
            throw {code: 1, message: CONSTANTS.ERROR_MESSAGES.ELEMENT_NOT_FOUND_CSS};
        } else {
            if (elements.length === 1) {
                return elements[0];
            }

            throw {code: 0, message: CONSTANTS.ERROR_MESSAGES.CSS_RETURNS_SEVERAL_ELEMENTS, elements: elements};
        }
    };

    const findElementInShadowDom = (css, targetDocument) => {
        let targetElement = targetDocument;
        let cssArray = css.split(' ');

        if (cssArray[cssArray.length - 1].includes(CONSTANTS.CSS.SHADOW_PSEUDO)) {
            throw new Error(CONSTANTS.ERROR_MESSAGES.SHADOW_CSS_CANNOT_END_ON_A_SHADOW_ROOT);
        }

        targetElement = cssArray.some(item => isSlotElement(item)) ?
            getTargetElementInSlottedContent(targetElement, cssArray) :
            getTargetElementInNotSlottedContent(targetElement, cssArray);

        if (!targetElement) {
            throw new Error(CONSTANTS.ERROR_MESSAGES.ELEMENT_NOT_FOUND_CSS);
        }

        return [targetElement];
    };

    const isSlotElement = element => {
        return element === CONSTANTS.TAGS.SLOT ||
            element.includes(CONSTANTS.TAGS.SLOT.concat('#')) ||
            element.includes(CONSTANTS.TAGS.SLOT.concat('.'));
    };

    const getTargetElementInSlottedContent = (targetElement, cssArray) => {
        cssArray.forEach(item => {
            let index = getIndexOfElementInSelector(item, getNodeName(targetElement) === CONSTANTS.TAGS.SLOT);
            item = getElementSelectorWithoutNthOfType(item);

            if (item.includes(CONSTANTS.CSS.SHADOW_PSEUDO)) {
                if (getNodeName(targetElement) === CONSTANTS.TAGS.SLOT) {
                    targetElement = getTargetElementFromSlot(targetElement, item, index);
                } else {
                    let selector = getSelectorForShadowRoot(item);
                    targetElement = getTargetElementBySelector(targetElement, selector, index);
                }

                if (!item.includes(CONSTANTS.CSS.SHADOW_PSEUDO_CLOSED)) {
                    targetElement = chrome.dom.openOrClosedShadowRoot(targetElement);
                }
            } else {
                if (getNodeName(targetElement) === CONSTANTS.TAGS.SLOT) {
                    targetElement = getTargetElementFromSlot(targetElement, item, index);
                } else {
                    targetElement = getTargetElementBySelector(targetElement, item, index);
                }
            }
        });

        return targetElement;
    };

    const getTargetElementInNotSlottedContent = (targetElement, cssArray) => {
        cssArray.forEach(item => {
            let index = getIndexOfElementInSelector(item);
            item = getElementSelectorWithoutNthOfType(item);

            if (item.includes(CONSTANTS.CSS.SHADOW_PSEUDO)) {
                let selector = getSelectorForShadowRoot(item);
                targetElement = getTargetElementBySelector(targetElement, selector, index);

                if (!item.includes(CONSTANTS.CSS.SHADOW_PSEUDO_CLOSED)) {
                    targetElement = chrome.dom.openOrClosedShadowRoot(targetElement);
                }
            } else {
                targetElement = getTargetElementBySelector(targetElement, item, index);
            }
        });

        return targetElement;
    };

    const getSelectorForShadowRoot = item => {
        return item.includes(CONSTANTS.CSS.SHADOW_PSEUDO_CLOSED) ?
            item.replace(CONSTANTS.CSS.SHADOW_PSEUDO_CLOSED, '') :
            item.replace(CONSTANTS.CSS.SHADOW_PSEUDO, '');
    };

    const getTargetElementFromSlot = (targetElement, selector, index) => {
        let assignedElements = targetElement.assignedElements();
        return index !== -1 ?
            assignedElements[index] :
            findTargetElementInSlot(selector.replace(CONSTANTS.CSS.SHADOW_PSEUDO, ''), assignedElements);
    };

    const getTargetElementBySelector = (targetElement, selector, index) => {
        return index !== -1 ?
            targetElement.querySelectorAll(selector)[index] :
            targetElement.querySelector(selector);
    };

    const findTargetElementInSlot = (selector, assignedElements) => {
        return assignedElements.find(element => element === element.parentElement.querySelector(selector));
    };

    const getIndexOfElementInSelector = (element, forSlot) => {
        const regex = new RegExp(CONSTANTS.CSS_HELPERS.NTH_PATTERN, 'gi');
        const resultArr = regex.exec(element);

        if (resultArr && resultArr[1]) {
            return resultArr[1].toString() - 1;
        }

        if (element.includes('#') || element.includes('.')) {
            return -1;
        }

        return forSlot ? -1 : 0;
    }

    const getElementSelectorWithoutNthOfType = element => {
        const regex = new RegExp(CONSTANTS.CSS_HELPERS.NTH_PATTERN, 'gi');
        const resultArr = regex.exec(element);
        if (resultArr && resultArr[0]) {
            return element.replace(resultArr[0], '');
        }

        return element;
    };

    return {
        generateCssForElementInShadowDom: generateCssForElementInShadowDom,
        getElementByCss: getElementByCss,
        getNodeName: getNodeName
    };
})();

subject7Gutenberg.locatorRecorder = (() => {
    let target = null;
    let originEvent = null;
    let targetWindow = null;
    let isTargetInShadowRoot = false;
    let hasTargetClosedShadowRoot = false;
    let targetComposedPath = null;
    let iframeDialogHint;
    let dialogHintContainer;
    const dialogHintBlockId = 'subject7GutenbergDlgH';
    let iframeDialogLocator;
    let dialogLocatorContainer;
    const dialogLocatorBlockId = 'subject7GutenbergDlgL';
    let lastTimeoutHandle = 0;
    let waitDblClickTimeoutId = 0;

    const generateSmartXpath = async (hint, callback) => {
        if (targetWindow !== window) {
            return;
        }

        const resultData = {
            hint
        };
        try {
            resultData.xpath = !isTargetInShadowRoot ?
                subject7Gutenberg.smartXpathEngine.generateSmartXPath(target, hint, targetWindow.document) :
                null;
        } catch (error) {
            showMessage(error.message);
            callback({hasError: true, message: error.message});
            return;
        }

        const frameXpathIfExist = await getIframeXpath(targetWindow);
        if (frameXpathIfExist !== null) {
            resultData.frame = frameXpathIfExist;
        }

        callback(resultData);
    };

    const getIframeXpath = async window => {
        const checkIsParentFrame = (paramWindow) => new Promise((resolve, reject) => {
            const channel = new MessageChannel();

            channel.port1.onmessage = event => {
                channel.port1.close();
                if (event.data.error) {
                    reject(event.data.error);
                } else {
                    resolve(event.data.result);
                }
            };

            paramWindow.parent.postMessage(
                {
                    sender: CONSTANTS.MESSAGE_SENDERS.CHILD_WINDOW,
                    action: CONSTANTS.ACTIONS.CHECK_IS_FRAME
                },
                '*',
                [channel.port2]
            );
        });

        let resultFrameAnalysis = getFrameInfo();
        let isParentFrame;
        if ((resultFrameAnalysis.crossOrigin && resultFrameAnalysis.sandboxed) ||
            (resultFrameAnalysis.crossOrigin && !resultFrameAnalysis.sandboxed) ||
            (!resultFrameAnalysis.crossOrigin && resultFrameAnalysis.sandboxed)) {
            if (resultFrameAnalysis.sandboxed && resultFrameAnalysis.sandboxAllowances &&
                (resultFrameAnalysis.sandboxAllowances.sameOrigin || resultFrameAnalysis.sandboxAllowances.scripts)) {
                isParentFrame = window.parent.location !== window.top.location;
            } else {
                isParentFrame = await checkIsParentFrame(window);
            }
        } else {
            isParentFrame = window.parent.location !== window.top.location;
        }

        if (isParentFrame) {
            let xpath = await getIframeXpath(window.parent);
            return xpath + ';' + await getFrameXpathIfExist(window);
        } else {
            return await getFrameXpathIfExist(window);
        }
    };

    const startCapturing = () => {
        target = null;
        originEvent = null;
        targetWindow = null;
        isTargetInShadowRoot = false;
        hasTargetClosedShadowRoot = false;
        targetComposedPath = null;
        blockUI(false);
        captureEvents(true);
        addDialogBlock(CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE);
        addDialogBlock(CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE);
        started = true;
    };

    const stopCapturing = () => {
        target = null;
        originEvent = null;
        targetWindow = null;
        isTargetInShadowRoot = false;
        hasTargetClosedShadowRoot = false;
        targetComposedPath = null;
        blockUI(false);
        captureEvents(false);
        removeDialogBlock(CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE);
        removeDialogBlock(CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE);
        started = false;
    };

    const resetCapturing = fireEvent => {
        if (targetWindow !== window) {
            return;
        }

        blockUI(false);
        if (fireEvent) {
            captureEvents(false);
            if (isTargetInShadowRoot) {
                targetComposedPath[0].dispatchEvent(originEvent);
            } else {
                target.dispatchEvent(originEvent);
            }
        }
        target = null;
        originEvent = null;
        targetWindow = null;
        isTargetInShadowRoot = false;
        targetComposedPath = null;
        captureEvents(started);
    };

    const test = (data, type, callback) => {
        if (targetWindow !== window) {
            return;
        }

        let element;
        try {
            element = type === CONSTANTS.TEST_TYPES.XPATH ?
                subject7Gutenberg.smartXpathEngine.getElementByXpath(data, targetWindow.document) :
                subject7Gutenberg.cssSelectorEngine.getElementByCss(data, targetWindow.document);
        } catch (error) {
            showMessage(error.message);
            callback();
            return;
        }
        scrollToElement(element);
        highlight(element, callback);
    };

    const blockUI = enabled => {
        const blockUiId = 'subject7GutenbergBlockUI';
        if (window !== window.top) {
            window.top.postMessage({
                sender: CONSTANTS.MESSAGE_SENDERS.CHILD_WINDOW,
                action: CONSTANTS.ACTIONS.BLOCK_UI,
                enabled
            }, '*');
            return;
        }

        const documentToBLock = window.frameElement ? window.top.document : window.document;
        let blockElement;
        if (enabled) {
            blockElement = documentToBLock.createElement(CONSTANTS.TAGS.DIV);
            blockElement.setAttribute(CONSTANTS.ATTRIBUTES.ID, blockUiId);
            blockElement.setAttribute(CONSTANTS.ATTRIBUTES.STYLE,
                'position: fixed; ' +
                'left: 0px; ' +
                'top: 0px; ' +
                'width: 100%; ' +
                'height: 100%; ' +
                'z-index: 2147483646; ' +
                'background: rgba(0, 0, 0, 0.5);'
            );
            documentToBLock.body.appendChild(blockElement);
        } else {
            blockElement = documentToBLock.getElementById(blockUiId);
            if (blockElement) {
                documentToBLock.body.removeChild(blockElement);
            }
        }
    };

    const captureEvents = enabled => {
        if (enabled) {
            window.addEventListener(CONSTANTS.EVENTS.DOM.CLICK, handleEvent, true);
            window.addEventListener(CONSTANTS.EVENTS.WINDOW.MESSAGE, handleIframeEvents, true);
        } else {
            window.removeEventListener(CONSTANTS.EVENTS.DOM.CLICK, handleEvent, true);
            window.removeEventListener(CONSTANTS.EVENTS.WINDOW.MESSAGE, handleIframeEvents, true);
        }
    };

    const getFrameInfo = () => {
        return sandblaster.detect() || {};
    };

    const handleIframeEvents = event => {
        const {data} = event;
        const {sender} = data;

        if (sender !== CONSTANTS.MESSAGE_SENDERS.CHILD_WINDOW) {
            return;
        }
        const {action} = data;
        switch (action) {
            case CONSTANTS.ACTIONS.BLOCK_UI:
                blockUI(data.enabled);
                break;
            case CONSTANTS.ACTIONS.GET_PRECISE_XPATH_FOR_FRAME:
                let iframe = getIFrame(getAllFrames(), data.iframeName, data.iframeSrc);
                let clarifiedXpath = null;
                if (iframe !== null) {
                    const idAttrValue = iframe.getAttribute(CONSTANTS.ATTRIBUTES.ID);
                    const nameAttrValue = iframe.getAttribute(CONSTANTS.ATTRIBUTES.NAME);
                    const titleAttrValue = iframe.getAttribute(CONSTANTS.ATTRIBUTES.TITLE);
                    let iframeXpath = CONSTANTS.XPATH_HELPERS.START_XPATH.concat(iframe.tagName).toLowerCase();

                    if (idAttrValue && idAttrValue.trim().length) {
                        clarifiedXpath = `${iframeXpath}[@id='${idAttrValue}']`;
                    } else if (nameAttrValue && nameAttrValue.trim().length) {
                        clarifiedXpath = `${iframeXpath}[@name='${nameAttrValue}']`;
                    } else if (titleAttrValue && titleAttrValue.trim().length) {
                        clarifiedXpath = `${iframeXpath}[@title='${titleAttrValue}']`;
                    }
                }
                event.ports[0].postMessage({result: clarifiedXpath});
                event.ports[0].close();
                break;
            case CONSTANTS.ACTIONS.CHECK_IS_FRAME:
                let resultFrameAnalysis = getFrameInfo();
                event.ports[0].postMessage({result: resultFrameAnalysis.framed});
                event.ports[0].close();
                break;
            case CONSTANTS.ACTIONS.CHECK_IS_FRAMESET:
                event.ports[0].postMessage({result: window.document.body.nodeName === CONSTANTS.TAGS.FRAMESET});
                event.ports[0].close();
                break;
            default:
                break;
        }
    };

    const getFrameXpathIfExist = async paramWindow => {
        let frame;
        let iframeXpath;
        let resultFrameAnalysis = getFrameInfo();

        if (!resultFrameAnalysis.framed) {
            return null;
        }

        if (resultFrameAnalysis.framed && !resultFrameAnalysis.crossOrigin) {
            try {
                frame = paramWindow.frameElement;
            } catch (e) {
                frame = null;
            }

            if (frame !== null) {
                const sandboxAttrValue = frame.getAttribute(CONSTANTS.ATTRIBUTES.SANDBOX);
                const idAttrValue = frame.getAttribute(CONSTANTS.ATTRIBUTES.ID);
                const nameAttrValue = frame.getAttribute(CONSTANTS.ATTRIBUTES.NAME);
                const srcAttrValue = frame.getAttribute(CONSTANTS.ATTRIBUTES.SRC);
                const titleAttrValue = frame.getAttribute(CONSTANTS.ATTRIBUTES.TITLE);

                iframeXpath = CONSTANTS.XPATH_HELPERS.START_XPATH.concat(frame.tagName).toLowerCase();

                if (idAttrValue && idAttrValue.trim().length) {
                    return `${iframeXpath}[@id='${idAttrValue}']`;
                }
                if (nameAttrValue && nameAttrValue.trim().length) {
                    return `${iframeXpath}[@name='${nameAttrValue}']`;
                }
                if (srcAttrValue && srcAttrValue.trim().length && srcAttrValue !== CONSTANTS.ATTRIBUTE_SRC_VALUE_SRCDOC) {
                    return `${iframeXpath}[@src='${srcAttrValue}']`;
                }
                if (titleAttrValue && titleAttrValue.trim().length) {
                    return `${iframeXpath}[@title='${titleAttrValue}']`;
                }
                if (resultFrameAnalysis.sandboxed && sandboxAttrValue) {
                    return sandboxAttrValue.trim().length ?
                        `${iframeXpath}[@sandbox='${sandboxAttrValue}']` :
                        `${iframeXpath}[@sandbox]`;
                }
            }

            return await getPreciseXpathForFrame(paramWindow, resultFrameAnalysis);
        }

        return await getPreciseXpathForFrame(paramWindow, resultFrameAnalysis);
    };

    const getPreciseXpathForFrame = async (paramWindow, resultFrameAnalysis) => {
        let iframeXpath = CONSTANTS.XPATH_HELPERS.START_XPATH + CONSTANTS.TAGS.IFRAME;
        try {
            const getPreciseXpath = (iframeName, iframeSrc) => new Promise((resolve, reject) => {
                const channel = new MessageChannel();

                channel.port1.onmessage = event => {
                    channel.port1.close();
                    if (event.data.error) {
                        reject(event.data.error);
                    } else {
                        resolve(event.data.result);
                    }
                };

                paramWindow.parent.postMessage(
                    {
                        sender: CONSTANTS.MESSAGE_SENDERS.CHILD_WINDOW,
                        action: CONSTANTS.ACTIONS.GET_PRECISE_XPATH_FOR_FRAME,
                        iframeSrc,
                        iframeName,
                    },
                    '*',
                    [channel.port2]
                );
            });

            let iframeSrc = paramWindow.location.href;
            let preciseIframeXpath = await getPreciseXpath(paramWindow.name, iframeSrc);
            if (preciseIframeXpath) {
                return preciseIframeXpath;
            }
            if (iframeSrc !== CONSTANTS.ATTRIBUTE_SRC_VALUE_SRCDOC) {
                return `${iframeXpath}[@src='${iframeSrc}']`
            }

            throw `iframe src attribute has value "${CONSTANTS.ATTRIBUTE_SRC_VALUE_SRCDOC}"`;
        } catch (e) {
            return resultFrameAnalysis.sandboxed ?
                getIframeXpathBySandboxAllowances(iframeXpath, resultFrameAnalysis.sandboxAllowances) :
                null;
        }
    };

    const getIframeXpathBySandboxAllowances = (iframeXpath, sandboxAllowances) => {
        if (sandboxAllowances) {
            if (sandboxAllowances.sameOrigin) {
                return `${iframeXpath}[contains(@sandbox, 'allow-same-origin')]`;
            }
            if (sandboxAllowances.forms) {
                return `${iframeXpath}[contains(@sandbox, 'allow-forms')]`;
            }
            if (sandboxAllowances.modals) {
                return `${iframeXpath}[contains(@sandbox, 'allow-modals')]`;
            }
            if (sandboxAllowances.orientationLock) {
                return `${iframeXpath}[contains(@sandbox, 'allow-orientation-lock')]`;
            }
            if (sandboxAllowances.pointerLock) {
                return `${iframeXpath}[contains(@sandbox, 'allow-pointer-lock')]`;
            }
            if (sandboxAllowances.popups) {
                return `${iframeXpath}[contains(@sandbox, 'allow-popups')]`;
            }
            if (sandboxAllowances.popupsToEscapeSandbox) {
                return `${iframeXpath}[contains(@sandbox, 'allow-popups-to-escape-sandbox')]`;
            }
            if (sandboxAllowances.presentation) {
                return `${iframeXpath}[contains(@sandbox, 'allow-presentation')]`;
            }
            if (sandboxAllowances.topNavigation) {
                return `${iframeXpath}[contains(@sandbox, 'allow-top-navigation')]`;
            }
            if (sandboxAllowances.topNavigationByUserActivation) {
                return `${iframeXpath}[contains(@sandbox, 'allow-top-navigation-by-user-activation')]`;
            }
        }

        return `${iframeXpath}[@sandbox]`;
    };

    const getAllFrames = () => {
        return document.getElementsByTagName(CONSTANTS.TAGS.IFRAME);
    };

    const getIFrame = (iframes, iframeName, iframeSrc) => {
        for (const f of iframes) {
            if ((iframeName.length !== 0 && f.name === iframeName) || iframeSrc === f.getAttribute(CONSTANTS.ATTRIBUTES.SRC)) {
                return f;
            }
        }

        return null;
    };

    const handleEvent = event => {
        if (checkIfTargetElementIsDialogFrame(event.target)) {
            return;
        }

        if (checkIfTargetElementIsHtmlOrBody(event.target)) {
            return;
        }

        defineShadowRootVariables(event.composedPath());
        target = findClickableElement(event.target);

        if (checkIfTargetElementIsHtmlOrBody(target)) {
            return;
        }

        originEvent = new event.constructor(event.type, event);
        targetWindow = window;
        event.preventDefault();
        event.stopPropagation();

        clearTimeout(waitDblClickTimeoutId);
        waitDblClickTimeoutId = setTimeout(async () => {
            captureEvents(false);
            blockUI(true);
            const isFrameset = window.frameElement ?
                await getIsFramesetProperty(window) :
                window.document.body.nodeName === CONSTANTS.TAGS.FRAMESET;
            chrome.runtime.sendMessage({
                action: CONSTANTS.ACTIONS.ELEMENT_CAPTURED,
                isFrameset,
                isDisabledHint: isTargetInShadowRoot || hasTargetClosedShadowRoot,
                isDisabledGenerating: hasTargetClosedShadowRoot
            });
        }, 320);
    };

    const isSVGElement = element => {
        return element instanceof SVGElement;
    };

    const getOpenOrClosedShadowRoot = element => {
        return isSVGElement(element) ?
            null :
            chrome.dom.openOrClosedShadowRoot(element);
    };

    const defineShadowRootVariables = composedPath => {
        targetComposedPath = composedPath;
        targetComposedPath.pop();
        targetComposedPath = targetComposedPath
            .filter(element => element.nodeType !== Node.DOCUMENT_NODE)
            .filter(element => subject7Gutenberg.cssSelectorEngine.getNodeName(element) !== CONSTANTS.TAGS.HTML)
            .filter(element => subject7Gutenberg.cssSelectorEngine.getNodeName(element) !== CONSTANTS.TAGS.BODY);

        hasTargetClosedShadowRoot = getOpenOrClosedShadowRoot(targetComposedPath[0]) !== null;

        isTargetInShadowRoot =
            !hasTargetClosedShadowRoot &&
            targetComposedPath.some(item => {
                return item.nodeType === Node.DOCUMENT_FRAGMENT_NODE || getOpenOrClosedShadowRoot(item) !== null
            });
    }

    const findClickableElement = element => {
        let parent = element;
        while (parent.parentNode !== null) {
            if (CONSTANTS.CLICKABLE_ELEMENT_TAGS.includes(parent.localName.toLowerCase())) {
                return parent;
            }
            parent = parent.parentNode;
        }
        return element;
    };

    const highlight = (element, callback) => {
        let count = 0;
        let originOutline = element.getAttribute(CONSTANTS.ATTRIBUTES.OUTLINE);
        let timer = setInterval(() => {
            if (count === 8) {
                clearInterval(timer);
                element.style.setProperty(CONSTANTS.ATTRIBUTES.OUTLINE, originOutline ? originOutline : '');
                callback();
                return;
            }
            if (count % 2 === 0) {
                element.style.setProperty(CONSTANTS.ATTRIBUTES.OUTLINE, '3px solid rgb(86, 135, 168)', 'important');
            } else {
                element.style.setProperty(CONSTANTS.ATTRIBUTES.OUTLINE, originOutline ? originOutline : '');
            }
            count++;
        }, 100);
    };

    const scrollToElement = element => {
        const elementRect = element.getBoundingClientRect();
        const absoluteElementTop = elementRect.top + window.pageYOffset;
        const middle = absoluteElementTop - (window.innerHeight / 2);
        window.scrollTo(0, middle);
    };

    const addDialogBlock = dialogType => {
        if (window.location !== window.parent.location) {
            return;
        }

        const documentToBLock = window.frameElement ? window.top.document : window.document;

        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE &&
            documentToBLock.getElementById(dialogHintBlockId)) {
            return;
        }

        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE &&
            documentToBLock.getElementById(dialogLocatorBlockId)) {
            return;
        }

        let dialogBlockId = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
            dialogHintBlockId : dialogLocatorBlockId;

        let dialogContainer = documentToBLock.createElement(CONSTANTS.TAGS.RECORDER_BLOCK);
        dialogContainer.setAttribute(CONSTANTS.ATTRIBUTES.ID, dialogBlockId);
        dialogContainer.setAttribute(CONSTANTS.ATTRIBUTES.CLASS,
            'draggable resizable '.concat(dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
                CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.CLASS_NAME : CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.CLASS_NAME)
        );
        let width = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
            CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.WIDTH : CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.WIDTH;
        dialogContainer.setAttribute(CONSTANTS.ATTRIBUTES.STYLE,
            'box-sizing: border-box; ' +
            'overflow: hidden; ' +
            'width: ' + width + 'px; ' +
            'visibility: hidden; ' +
            'z-index: 2147483647;' +
            'border-radius: 5px 0 0 5px; ' +
            'background: white; ' +
            'display: block; '
        );

        let handle = documentToBLock.createElement(CONSTANTS.TAGS.DIV);
        handle.setAttribute(CONSTANTS.ATTRIBUTES.ID, dialogBlockId.concat(CONSTANTS.DIALOG_ELEMENTS.HANDLE.LABEL));
        handle.setAttribute(CONSTANTS.ATTRIBUTES.CLASS,
            'handle '.concat(dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
                CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.CLASS_NAME : CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.CLASS_NAME)
        );
        handle.setAttribute(CONSTANTS.ATTRIBUTES.STYLE,
            'box-sizing: border-box; ' +
            'width: 30px; ' +
            'border: none; ' +
            'border-right: 1px solid #e3e3e3; ' +
            'height: 100%; ' +
            'position: absolute; ' +
            'top: 0; ' +
            'left: 0; ' +
            'display: flex; ' +
            'flex-direction: column; ' +
            'justify-content: center; ' +
            'align-items: center; ' +
            'cursor: grab; ' +
            'z-index: 1; '
        );

        let headerText = documentToBLock.createElement(CONSTANTS.TAGS.LABEL);
        headerText.setAttribute(CONSTANTS.ATTRIBUTES.STYLE,
            'font-size: 24px; ' +
            'font-weight: 400; ' +
            'font-family: Helvetica Neue,Helvetica,Arial,sans-serif;' +
            'color: white; ' +
            'padding-left: 2px; ' +
            'text-shadow: -1px -1px #aaaaaa; ' +
            'pointer-events: none;'
        );
        headerText.innerHTML = '⣿';
        handle.append(headerText);

        dialogContainer.append(handle);

        let iframeDialog = documentToBLock.createElement(CONSTANTS.TAGS.IFRAME);
        iframeDialog.setAttribute(CONSTANTS.ATTRIBUTES.ID, dialogBlockId.concat(CONSTANTS.DIALOG_ELEMENTS.IFRAME.LABEL));
        iframeDialog.setAttribute(CONSTANTS.ATTRIBUTES.SRC,
            chrome.runtime.getURL(dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
                CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.URL :
                CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.URL
            )
        );
        iframeDialog.setAttribute(CONSTANTS.ATTRIBUTES.STYLE,
            'box-sizing: border-box; ' +
            'width: calc(100% - 30px); ' +
            'top: auto; ' +
            'left: auto; ' +
            'margin-left: 30px; ' +
            'vertical-align: top;' +
            'border: none; '
        );
        dialogContainer.append(iframeDialog);

        documentToBLock.body.prepend(dialogContainer);

        dialogContainer.style.position = 'fixed';

        let draggableOptions = Object.assign({}, CONSTANTS.DRAGGABLE_OPTIONS);
        switch (dialogType) {
            case CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE:
                dialogHintContainer = dialogContainer;
                iframeDialogHint = iframeDialog;
                draggableOptions.handle = draggableOptions.handle.concat('.').concat(CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.CLASS_NAME);
                $('.draggable.resizable.s7-hint').draggable(draggableOptions);
                break;
            case CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE:
                dialogLocatorContainer = dialogContainer;
                iframeDialogLocator = iframeDialog;
                draggableOptions.handle = draggableOptions.handle.concat('.').concat(CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.CLASS_NAME);
                $('.draggable.resizable.s7-locator').draggable(draggableOptions);
                break;
            default:
                break;
        }
    };

    const removeDialogBlock = dialogType => {
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE && !dialogHintContainer) {
            return;
        }

        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE && !dialogLocatorContainer) {
            return;
        }

        switch (dialogType) {
            case CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE:
                dialogHintContainer.parentElement.removeChild(dialogHintContainer);
                dialogHintContainer = undefined;
                break;
            case CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE:
                dialogLocatorContainer.parentElement.removeChild(dialogLocatorContainer);
                dialogLocatorContainer = undefined;
                break;
            default:
                break;
        }
    };

    const openDialog = (dialogType, options) => {
        const {hint, locator, saveFeatureEnabled, isFrameset, isDisabledHint, isDisabledGenerating} = options;

        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE && !dialogHintContainer) {
            return;
        }
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE && !dialogLocatorContainer) {
            return;
        }

        const iframeDialog = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
            iframeDialogHint :
            iframeDialogLocator;

        clearTimeout(lastTimeoutHandle);
        closeDialog(dialogType);

        lastTimeoutHandle = setTimeout(
            () => {
                let message = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
                    {
                        sender: CONSTANTS.MESSAGE_SENDERS.S7_OPEN_HINT_WINDOW,
                        hint,
                        isFrameset,
                        isDisabledHint,
                        isDisabledGenerating
                    } :
                    {
                        sender: CONSTANTS.MESSAGE_SENDERS.S7_OPEN_LOCATOR_WINDOW,
                        hint,
                        locator,
                        saveFeatureEnabled,
                        isFrameset,
                        isDisabledHint
                    };
                iframeDialog.contentWindow.postMessage(message, '*');
            },
            dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
                CONSTANTS.OPEN_DIALOG_DELAY.HINT :
                CONSTANTS.OPEN_DIALOG_DELAY.LOCATOR
        );
    };

    const syncDialogSize = ({dialogType, dialogWidth, dialogHeight}) => {
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE && !dialogHintContainer) {
            return;
        }
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE && !iframeDialogLocator) {
            return;
        }

        const iframeDialog = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
            iframeDialogHint :
            iframeDialogLocator;
        const dialogContainer = dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE ?
            dialogHintContainer :
            dialogLocatorContainer;

        iframeDialog.style.height = Math.round(dialogHeight) + 'px';
        iframeDialog.style.width = Math.round(dialogWidth) + 'px';

        const dialogContainerRect = dialogContainer.getBoundingClientRect();
        const dialogContainerHeight = dialogContainerRect.height;
        const dialogContainerWidth = dialogContainerRect.width;

        const pageHeight = document.documentElement.clientHeight;
        const pageWidth = document.documentElement.clientWidth;
        let dialogTop = 0;
        let dialogLeft = 0;
        if (originEvent && originEvent.clientX && originEvent.clientY) {
            dialogTop = originEvent.clientY;
            if (dialogTop + dialogContainerHeight > pageHeight) {
                dialogTop = Math.max(0, originEvent.clientY - dialogContainerHeight);
            }
            dialogLeft = originEvent.clientX;
            if (dialogLeft + dialogContainerWidth > pageWidth) {
                dialogLeft = Math.max(0, originEvent.clientX - dialogContainerWidth);
            }
        } else {
            dialogTop = pageHeight / 2 - dialogContainerHeight / 2;
            dialogLeft = pageWidth / 2 - dialogContainerWidth / 2;
        }
        dialogContainer.style.top = Math.round(dialogTop) + 'px';
        dialogContainer.style.left = Math.round(dialogLeft) + 'px';
        dialogContainer.style.visibility = 'visible';
    };

    const closeDialog = dialogType => {
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE) {
            if (!dialogHintContainer) {
                return;
            }
            if (dialogHintContainer.style.visibility === 'visible') {
                dialogHintContainer.style.visibility = 'hidden';
            }
        }
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE) {
            if (!dialogLocatorContainer) {
                return;
            }
            if (dialogLocatorContainer.style.visibility === 'visible') {
                dialogLocatorContainer.style.visibility = 'hidden';
            }
        }
    };

    const setDialogVisible = (dialogType, isVisible) => {
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.HINT.TYPE) {
            if (!dialogHintContainer) {
                return;
            }
            dialogHintContainer.style.visibility = isVisible ? 'visible' : 'hidden';
        }
        if (dialogType === CONSTANTS.DIALOG_FRAME_OPTIONS.LOCATOR.TYPE) {
            if (!dialogLocatorContainer) {
                return;
            }
            dialogLocatorContainer.style.visibility = isVisible ? 'visible' : 'hidden';
        }
    };

    const checkIfTargetElementIsDialogFrame = target => {
        return target.id === dialogHintBlockId.concat(CONSTANTS.DIALOG_ELEMENTS.HANDLE.LABEL) ||
            target.id === dialogLocatorBlockId.concat(CONSTANTS.DIALOG_ELEMENTS.HANDLE.LABEL)
    };

    const checkIfTargetElementIsHtmlOrBody = target => {
        if (!target.tagName) {
            return false;
        }

        return getTagName(target) === CONSTANTS.TAGS.HTML || getTagName(target) === CONSTANTS.TAGS.BODY;
    };

    const getTagName = target => {
        return target.tagName.toLowerCase();
    };

    const showMessage = string => {
        if (navigator.webdriver) {
            chrome.runtime.sendMessage({
                action: CONSTANTS.ACTIONS.SEND_NOTIFICATION,
                id: CONSTANTS.NOTIFICATION_IDS.CONTENT,
                message: string
            });
        } else {
            alert(string);
        }
    };

    const getIsFramesetProperty = async window => {
        const checkIsFrameset = (paramWindow) => new Promise((resolve, reject) => {
            const channel = new MessageChannel();

            channel.port1.onmessage = event => {
                channel.port1.close();
                if (event.data.error) {
                    reject(event.data.error);
                } else {
                    resolve(event.data.result);
                }
            };

            paramWindow.top.postMessage(
                {
                    sender: CONSTANTS.MESSAGE_SENDERS.CHILD_WINDOW,
                    action: CONSTANTS.ACTIONS.CHECK_IS_FRAMESET
                },
                '*',
                [channel.port2]
            );
        });

        return await checkIsFrameset(window);
    };

    const generateCssForElementInShadowDom = () => {
        try {
            return subject7Gutenberg.cssSelectorEngine.generateCssForElementInShadowDom(targetComposedPath);
        } catch (error) {
            showMessage(error.message);
            return null;
        }
    };

    return {
        generateSmartXpath: generateSmartXpath,
        generateCssForElementInShadowDom: generateCssForElementInShadowDom,
        resetCapturing: resetCapturing,
        startCapturing: startCapturing,
        stopCapturing: stopCapturing,
        test: test,
        openDialog: openDialog,
        closeDialog: closeDialog,
        syncDialogSize: syncDialogSize,
        setDialogVisible: setDialogVisible,
        getIsFramesetProperty: getIsFramesetProperty,
        getFrameInfo: getFrameInfo,
        getOpenOrClosedShadowRoot: getOpenOrClosedShadowRoot
    };

})();

chrome.runtime.onMessage.addListener(async (request, sender, sendResponse) => {
    let action = request.action;

    if (action === CONSTANTS.ACTIONS.CHECK_IS_AUTOMATION_BROWSER) {
        sendResponse({isAutomationBrowser: navigator.webdriver});
        return true;
    }

    switch (action) {
        case CONSTANTS.ACTIONS.START_CAPTURING:
            subject7Gutenberg.locatorRecorder.startCapturing();
            sendResponse({started: started});
            break;
        case CONSTANTS.ACTIONS.STOP_CAPTURING:
            subject7Gutenberg.locatorRecorder.stopCapturing();
            sendResponse({started: started});
            break;
        case CONSTANTS.ACTIONS.TEST_XPATH:
            if (request.xpath) {
                subject7Gutenberg.locatorRecorder.test(request.xpath, CONSTANTS.TEST_TYPES.XPATH, () => {
                    chrome.runtime.sendMessage({
                        action: CONSTANTS.ACTIONS.TEST_XPATH_FINISHED,
                        css: request.css,
                        isFrameset: request.isFrameset
                    });
                });
            } else {
                chrome.runtime.sendMessage({
                    action: CONSTANTS.ACTIONS.TEST_XPATH_FINISHED,
                    css: request.css,
                    isFrameset: request.isFrameset
                });
            }
            break;
        case CONSTANTS.ACTIONS.TEST_CSS:
            if (request.css) {
                subject7Gutenberg.locatorRecorder.test(request.css, CONSTANTS.TEST_TYPES.CSS, () => {
                    chrome.runtime.sendMessage({
                        action: CONSTANTS.ACTIONS.TEST_CSS_FINISHED,
                        isFrameset: request.isFrameset
                    });
                });
            } else {
                chrome.runtime.sendMessage({
                    action: CONSTANTS.ACTIONS.TEST_CSS_FINISHED,
                    isFrameset: request.isFrameset
                });
            }
            break;
        case CONSTANTS.ACTIONS.GENERATE_SMART_XPATH:
            const {hint, isDisabledHint} = request;
            await subject7Gutenberg.locatorRecorder.generateSmartXpath(hint, async result => {
                const isFrameset = window.frameElement ?
                    await subject7Gutenberg.locatorRecorder.getIsFramesetProperty(window) :
                    window.document.body.nodeName === CONSTANTS.TAGS.FRAMESET;
                result.css = !result.xpath ?
                    subject7Gutenberg.locatorRecorder.generateCssForElementInShadowDom() :
                    null;
                chrome.runtime.sendMessage({
                    action: CONSTANTS.ACTIONS.GENERATION_FINISHED,
                    result,
                    hint,
                    isFrameset,
                    isDisabledHint
                });
            });
            break;
        case CONSTANTS.ACTIONS.RESET_CAPTURING:
            subject7Gutenberg.locatorRecorder.resetCapturing(request.fireEvent);
            break;
        case CONSTANTS.ACTIONS.OPEN_DIALOG:
            const {dialogType, options} = request;
            subject7Gutenberg.locatorRecorder.openDialog(dialogType, options);
            break;
        case CONSTANTS.ACTIONS.CLOSE_DIALOG:
            subject7Gutenberg.locatorRecorder.closeDialog(request.dialogType);
            break;
        case CONSTANTS.ACTIONS.SYNC_DIALOG_SIZE:
            subject7Gutenberg.locatorRecorder.syncDialogSize(request);
            break;
        case CONSTANTS.ACTIONS.HIDE_DIALOG:
            subject7Gutenberg.locatorRecorder.setDialogVisible(request.dialogType, false);
            break;
        case CONSTANTS.ACTIONS.SHOW_DIALOG:
            subject7Gutenberg.locatorRecorder.setDialogVisible(request.dialogType, true);
            break;
        case CONSTANTS.ACTIONS.SHOW_ALERT:
            if (window === window.top) {
                alert(request.message);
            }
            break;
        default:
            console.debug(`${new Date(Date.now()).toISOString()} -> Action is not defined: ${JSON.stringify(request)}`);
            break;
    }
    sendResponse();
    return true;
});

let port;
const connect = () => {
    port = chrome.runtime.connect(chrome.runtime.id, {name: CONSTANTS.MESSAGING_PORT});
    console.debug(`${new Date(Date.now()).toISOString()} -> [KEEP_ALIVE] Content script starts connection to port '${port.name}'`);
    port.onDisconnect.addListener(connect);
    port.onMessage.addListener(msg => {
        console.debug(`${new Date(Date.now()).toISOString()} -> [MESSAGE_EXCHANGE] Received ${JSON.stringify(msg)} from Service Worker`);
    });
    port.postMessage({isAutomationBrowser: navigator.webdriver});
};

if (!subject7Gutenberg.locatorRecorder.getFrameInfo().framed) {
    connect();
}
